/* HTTP data interpreter.
   Copyright (C) 2011 Sergey Poznyakoff

   JOH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   JOH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JOH.  If not, see <http://www.gnu.org/licenses/>. */

#include "joh.h"

enum sep_state {
	sep_init,
	sep_cr1,
	sep_lf1,
	sep_cr2,
	sep_lf2
};

struct http_decode_state {
	enum sep_state state;
	size_t off;
	size_t content_length;
};

static struct http_decode_state *
get_state(void **pstate)
{
	struct http_decode_state *p;
	if (!pstate)
		return NULL;
	if (!*pstate) {
		p = xzalloc(sizeof(*p));
		*pstate = p;
	} else
		p = *pstate;
	return p;
}

static int transtab[][255] = {
	{  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, },
	{  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  2,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, },
	{  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  3,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, },
	{  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  4,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, }
};

enum sep_state 
find_separator(enum sep_state state,
	       const unsigned char *iptr, size_t isize,
	       size_t off, size_t *poff)
{
	iptr += off;
	isize -= off;

	for (; isize && state != sep_lf2; iptr++, isize--, off++)
		state = transtab[state][*iptr];
	
	*poff = off;
	return state;
}

static char *
find_eol(const unsigned char *text)
{
	enum sep_state state = sep_init;
	
	for (; *text && state != sep_lf1; text++)
		state = transtab[state][*text];
	return (char*)text;
}


#define ISWS(c) ((c)==' '||(c)=='\t')
#define SKIPWS(p,id)						\
	for (; ISWS(*p); p++);					\
	if (!*p) {						\
		error("%s: malformed POST request", id);	\
		return -1;					\
	}							\

static int
parse_post_request(char *ptr, struct joh_param param[],
		   struct http_decode_state *st,
		   const char *id)
{
	char *query;

	/* 1. Parse POST line */
	if (strncasecmp(ptr, "POST", 4)) {
		error("%s: not a POST request", id);
		return -1;
	}
	ptr += 4;
	if (!ISWS(*ptr)) {
		error("%s: malformed POST request", id);
		return -1;
	}
	SKIPWS(ptr, id);
	
	query = ptr;
	for (; !ISWS(*ptr); ptr++);
	if (*ptr == 0) {
		error("%s: malformed POST request", id);
		return -1;
	}
	*ptr++ = 0;
	
	debug(50,("query=%s",query));
	SKIPWS(ptr, id);

	if (strncmp(ptr, "HTTP/", 5)) {
		error("%s: malformed POST request", id);
		return -1;
	}
	ptr += 5;

	ptr = find_eol((unsigned char *)ptr);
	if (!*ptr)
		return -1;

	query = strchr(query, '?');
	if (*query) {
		parse_http_query(param, query + 1);
		debug(50,("server=%s",param[joh_param_server].ptr));
		debug(50,("port=%s",param[joh_param_port].ptr));
	}

		       
	/* 2. Scan headers */
#define CONTENT_LENGTH_HEADER "Content-Length:"
#define CONTENT_LENGTH_HEADER_LEN (sizeof(CONTENT_LENGTH_HEADER)-1)
#define CONTENT_TYPE_HEADER "Content-Type:"
#define CONTENT_TYPE_HEADER_LEN (sizeof(CONTENT_TYPE_HEADER)-1)
#define CONTENT_TYPE "application/x-www-form-urlencoded"
#define CONTENT_TYPE_LEN (sizeof(CONTENT_TYPE)-1)
	
	while (*ptr) {
		if (strncasecmp(ptr, CONTENT_LENGTH_HEADER,
				CONTENT_LENGTH_HEADER_LEN) == 0) {
			char *p;

			st->content_length =
				strtoul(ptr + CONTENT_LENGTH_HEADER_LEN,
					&p, 10);
			for (;*p && ISWS(*p); p++);
			if (*p != '\r')
				return -1;
			ptr = p;
		} else if (strncasecmp(ptr, CONTENT_TYPE_HEADER,
				       CONTENT_TYPE_HEADER_LEN) == 0) {
			for (ptr += CONTENT_TYPE_HEADER_LEN; *ptr && ISWS(*ptr);
			     ptr++);
			if (strncmp(ptr, CONTENT_TYPE, CONTENT_TYPE_LEN))
				return -1;
		}
		ptr = find_eol((unsigned char*)ptr);
	}
	
	return 0;
}

int
joh_http_decode(void **pstate, struct joh_param param[joh_param_max],
		joh_cbuf_t ibuf, const char *id)
{
	struct http_decode_state *st = get_state(pstate);
	size_t isize;
	char *iptr;

	debug(50, ("joh_http_decode called"));
	joh_cbuf_coalesce(ibuf);
	iptr = joh_cbuf_head(ibuf, &isize);

	if (st->state != sep_lf2) {
		st->state = find_separator(st->state,
					   (unsigned char*)iptr, isize,
					   st->off, &st->off);
		if (st->state != sep_lf2)
			return 1;
		debug(50, ("Found separator at %lu", (unsigned long) st->off));

		iptr[st->off-2] = 0;
		if (parse_post_request(iptr, param, st, id))
			return -1;
	}

	if (st->content_length == 0)
		return 0;

	if (isize < st->off + st->content_length)
		return 1;

	if (parse_content(param, iptr + st->off, st->content_length))
		abort();//FIXME
	free(st);
	*pstate = NULL;

	if (debug_level >= 50) {
		int i;

		for (i = 0; i < joh_param_max; i++)
			debug(50, ("%s: %s=%s", id, param_str[i],
				   param[i].ptr));
	}
	return 0;
}

/* Format HTTP reply */

static void
fmt_string(joh_cbuf_t tbuf, const char *str,
	   struct joh_param *param JOH_ARG_UNUSED)
{
	joh_cbuf_append(tbuf, str, strlen(str));
}

static void
fmt_server(joh_cbuf_t tbuf, const char *str,
	   struct joh_param *param JOH_ARG_UNUSED)
{
	joh_cbuf_append(tbuf, str, strlen(str));
	joh_cbuf_append_char(tbuf, ' ');
	fmt_string(tbuf, PACKAGE_NAME "/" PACKAGE_VERSION, param);
}

static void
fmt_cookie(joh_cbuf_t tbuf, const char *str, struct joh_param *param)
{
	if (param[joh_param_id].ptr) {
		joh_cbuf_append(tbuf, str, strlen(str));
		joh_cbuf_append(tbuf, " ID=", 4);
		joh_cbuf_append(tbuf, param[joh_param_id].ptr,
				param[joh_param_id].size);
	}
}

static void
fmt_date(joh_cbuf_t tbuf, const char *str,
	   struct joh_param *param JOH_ARG_UNUSED)
{
	char datebuf[32];
	time_t t = time(NULL);
	struct tm *tm;
	
	joh_cbuf_append(tbuf, str, strlen(str));
	joh_cbuf_append_char(tbuf, ' ');
	tm = gmtime(&t);
	strftime(datebuf, sizeof(datebuf), "%a, %d %b %Y %H:%M:%S GMT", tm);
	joh_cbuf_append(tbuf, datebuf, strlen(datebuf));
}

static void
fmt_content_length(joh_cbuf_t tbuf, const char *str, struct joh_param *param)
{
	size_t size = param[joh_param_payload].size;
	char nbuf[128];

	/* FIXME */
	snprintf(nbuf, sizeof(nbuf), "%lu", (unsigned long) size);
	joh_cbuf_append(tbuf, str, strlen(str));
	joh_cbuf_append_char(tbuf, ' ');
	joh_cbuf_append(tbuf, nbuf, strlen(nbuf));
}

struct reply_fmt {
	const char *text;
	void (*fun)(joh_cbuf_t tbuf, const char *str, struct joh_param *param);
};

struct reply_fmt reply_tab[] = {
	{ "HTTP/1.1 200 OK", fmt_string },
	{ "Date:", fmt_date },
	{ "Server:", fmt_server },
	{ "Set-Cookie:", fmt_cookie },
	{ "Connection: Close", fmt_string },
	{ "Content-Type: text/xml", fmt_string },
	{ "Content-Length: ", fmt_content_length },
	{ NULL }
};

static void
http_response(struct joh_obuf *obuf, struct joh_param *param)
{
	joh_cbuf_t tbuf = joh_cbuf_create();
	struct reply_fmt *fmt;
	char *p;
	size_t size;
	
	for (fmt = reply_tab; fmt->fun; fmt++) {
		fmt->fun(tbuf, fmt->text, param);
		joh_cbuf_append(tbuf, "\r\n", 2);
	}
	joh_cbuf_append(tbuf, "\r\n", 2);
	if (param[joh_param_payload].ptr) {
		joh_cbuf_append(tbuf, param[joh_param_payload].ptr,
				param[joh_param_payload].size);
	}
	size = joh_cbuf_size(tbuf);
	p = joh_cbuf_finish(tbuf);
	obuf->base = xmalloc(size);
	memcpy(obuf->base, p, size);
	obuf->base_alloc = 1;
	obuf->size = size;
	obuf->off = 0;
	joh_cbuf_free(tbuf);
}

int
joh_http_encode(struct joh_obuf *obuf, size_t count, const int *param_idx,
		struct joh_param *param_vals)
{
	size_t i;
	struct joh_param param[joh_param_max];

	memset(param, 0, sizeof(param));
	for (i = 0; i < count; i++)
		param[param_idx[i]] = param_vals[i];
	http_response(obuf, param);
	return 0;
}
